/*
 *  fair_share.c - A simple weight based Thermal governor
 *
 *  Copyright (C) 2012 Intel Corp
 *  Copyright (C) 2012 Durgadoss R <durgadoss.r@intel.com>
 *
 *  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */

#include <linux/thermal.h>
#include <linux/slab.h>
#include <trace/events/thermal.h>

#include "thermal_core.h"

struct avm_fair_share_data {
	int current_trip;
};

/**
 * get_trip_level: - obtains the current trip level for a zone
 * @tz:		thermal zone device
 */
static int get_trip_level(struct thermal_zone_device *tz)
{
	int trip = 0;
	int trip_temp, trip_hyst;
	enum thermal_trip_type trip_type;
	struct avm_fair_share_data *data = tz->governor_data;

	if (tz->trips == 0 || !tz->ops->get_trip_temp)
		return 0;

	trip = data->current_trip;

	while (trip >= -1 && trip < tz->trips) {
		// Check if we need to raise the trip level
		if (trip + 1 < tz->trips) {
			tz->ops->get_trip_temp(tz, trip + 1, &trip_temp);

			// Raise it and try again
			if (tz->temperature >= trip_temp) {
				trip++;
				continue;
			}
		}

		if (trip >= 0) {
			// Check if we need to lower the trip and try agian
			tz->ops->get_trip_temp(tz, trip, &trip_temp);
			tz->ops->get_trip_hyst(tz, trip, &trip_hyst);

			if (tz->temperature < trip_temp - trip_hyst) {
				trip--;
				continue;
			}
		}

		break;
	}

	/*
	 * count > 0 only if temperature is greater than first trip
	 * point, in which case, trip_point = count - 1
	 */
	if (trip >= 0) {
		tz->ops->get_trip_type(tz, trip, &trip_type);
		trace_thermal_zone_trip(tz, trip, trip_type);
	}

	if (data->current_trip != trip)
		pr_err("[avm_fair_share] change trip level %d -> %d for temp %d\n",
		       data->current_trip, trip, tz->temperature);
	data->current_trip = trip;

	return trip + 1;
}

static long get_target_state(struct thermal_zone_device *tz,
		struct thermal_cooling_device *cdev, int percentage, int level)
{
	unsigned long max_state;

	cdev->ops->get_max_state(cdev, &max_state);

	return (long)(percentage * level * max_state) / (100 * tz->trips);
}

/**
 * fair_share_throttle - throttles devices associated with the given zone
 * @tz - thermal_zone_device
 *
 * Throttling Logic: This uses three parameters to calculate the new
 * throttle state of the cooling devices associated with the given zone.
 *
 * Parameters used for Throttling:
 * P1. max_state: Maximum throttle state exposed by the cooling device.
 * P2. percentage[i]/100:
 *	How 'effective' the 'i'th device is, in cooling the given zone.
 * P3. cur_trip_level/max_no_of_trips:
 *	This describes the extent to which the devices should be throttled.
 *	We do not want to throttle too much when we trip a lower temperature,
 *	whereas the throttling is at full swing if we trip critical levels.
 *	(Heavily assumes the trip points are in ascending order)
 * new_state of cooling device = P3 * P2 * P1
 */
static int fair_share_throttle(struct thermal_zone_device *tz, int trip)
{
	struct thermal_instance *instance;
	int total_weight = 0;
	int total_instance = 0;
	int cur_trip_level = get_trip_level(tz);

	list_for_each_entry(instance, &tz->thermal_instances, tz_node) {
		if (instance->trip != trip)
			continue;

		total_weight += instance->weight;
		total_instance++;
	}

	list_for_each_entry(instance, &tz->thermal_instances, tz_node) {
		int percentage;
		struct thermal_cooling_device *cdev = instance->cdev;

		if (instance->trip != trip)
			continue;

		if (!total_weight)
			percentage = 100 / total_instance;
		else
			percentage = (instance->weight * 100) / total_weight;

		instance->target = get_target_state(tz, cdev, percentage,
						    cur_trip_level);

		instance->cdev->updated = false;
		thermal_cdev_update(cdev);
	}
	return 0;
}

static int avm_fair_share_bind(struct thermal_zone_device *tz)
{
	tz->governor_data = kzalloc(sizeof(struct avm_fair_share_data), GFP_KERNEL);

	return 0;
}

static void avm_fair_share_unbind(struct thermal_zone_device *tz)
{
	kfree(tz->governor_data);
}

static struct thermal_governor thermal_gov_fair_share = {
	.name		= "fair_share",
	.bind_to_tz	= avm_fair_share_bind,
	.unbind_from_tz = avm_fair_share_unbind,
	.throttle	= fair_share_throttle,
};

int thermal_gov_fair_share_register(void)
{
	return thermal_register_governor(&thermal_gov_fair_share);
}

void thermal_gov_fair_share_unregister(void)
{
	thermal_unregister_governor(&thermal_gov_fair_share);
}

